from enum import Enum


class Tier(str, Enum):
    chall = "challenger"
    gm = "grandmaster"
    m = "master"
    d = "diamond"
    e = "emerald"
    p = "platinum"
    g = "gold"
    s = "silver"
    b = "bronze"
    i = "iron"
    m_ = "master_plus"
    d_ = "diamond_plus"
    e_ = "emerald_plus"
    p_ = "platinum_plus"
    g_ = "gold_plus"



from abc import ABC, abstractmethod
from pathlib import Path

from helpmepick.models.champion import Champion

class BaseChampIO(ABC):

    def __init__(self, data_folder: Path) -> None:
        assert data_folder.exists(), f"Passed data folder {data_folder} does not exist"
        assert data_folder.is_dir(), f"Passed data folder {data_folder} is not a directory"
        self.root = data_folder
    
    @abstractmethod
    def get_champ_path(self, champion: Champion) -> Path:
        pass

    @abstractmethod
    def load_champ(self, champion: Champion) -> list[tuple[Champion, float, int]]:
        pass

    @abstractmethod
    def save_champ(self, champion: Champion, data: list[tuple[Champion, float, int]]) -> None:
        pass



from enum import EnumMeta
from pathlib import Path
from helpmepick.models.champion import Champion
from helpmepick.models.region import Region
from helpmepick.models.role import Role, ALL_ROLE_NAMES
from helpmepick.models.tier import Tier
from helpmepick.models.pool import Pool
from helpmepick.persistence.csv import CsvChampIO
from helpmepick.pool.yaml import YamlPoolIO
from helpmepick.utils.throttler import Throttler
from helpmepick.services.opgg import retrieve_opgg_data, parse_data

from typing import Annotated, Callable, Iterable
import typer

DEFAULT_POOL_FNAME = "./pool.yaml"
DEFAULT_DATA_FOLDER = "./data"


def completer(cls: EnumMeta) -> Callable[[str], Iterable[str]]:
    def complete_f(x: str) -> Iterable[str]:
        for elem in cls:
            if elem.value.startswith(x):
                yield elem

    return complete_f


role_cli = Annotated[
    Role, typer.Argument(help="Selected Role", autocompletion=completer(Role))
]

champ_cli = Annotated[
    Champion, typer.Argument(help="Selected Champion", autocompletion=completer(Champion))
]

app = typer.Typer()
poolIO = YamlPoolIO(Path(DEFAULT_POOL_FNAME))

@app.command()
def init():
    # TODO: de-hardcode this, move onto config file
    pool_fname = DEFAULT_POOL_FNAME
    data_folder = DEFAULT_DATA_FOLDER

    pool_path = Path(pool_fname)
    if not pool_path.exists():
        print("Creating pool file at", str(pool_path))
        #pool_path.touch()
        poolIO.save_pool(Pool({Role(role):{} for role in ALL_ROLE_NAMES}))

    data_path = Path(data_folder)
    if not data_path.exists():
        print("Creating data folder at", str(data_path))
        data_path.mkdir()
        for role in ALL_ROLE_NAMES:
            (data_path / role).mkdir()


@app.command()
def add(
    role: role_cli,
    champion: champ_cli,
    priority: Annotated[
        int, typer.Argument(help="Priority for this champion, 1 is maximum")
    ],
):
    pool = poolIO.load_pool()
    pool.add_champion(champion, role, priority)
    poolIO.save_pool(pool)
    print(
        f"Champion {champion} successfully added to {role} role pool with priority {priority}"
    )

@app.command()
def update(role: role_cli):
    # TODO: add params for throttling, tiers, allow all roles, etc.
    pool = poolIO.load_pool()
    root_path = Path(DEFAULT_DATA_FOLDER)
    print("Updating role", role.name)
    champIO = CsvChampIO(root_path / role.name)
    for champ in Throttler(pool.get_champions_for(Role(role))):
        print("Updating champion", champ.name, end="...\t")
        resp = retrieve_opgg_data(champ, role)
        data = parse_data(resp)
        champIO.save_champ(champ,data)
        print("OK")

@app.command("vs")
def recommend(role: role_cli, enemy_champion: champ_cli):
    # TODO: add support for champions in pool where enemy is not represented
    rolepool = poolIO.load_pool().get_champions_for(role)
    root_path = Path(DEFAULT_DATA_FOLDER)
    champIO = CsvChampIO(root_path / role.name)
    matchup_data = []
    no_data = []
    # compile data
    for champion, priority in rolepool.items():
        data_dict = {
            champ:(winrate, games) 
            for champ, winrate, games 
            in champIO.load_champ(champion)
        }
        if enemy_champion in data_dict:
            matchup_data.append(
                (champion, priority, *data_dict[enemy_champion])
            )
        else:
            no_data.append((champion, priority))
    # print it sorted by winrate>priority>games
    for champion, priority, winrate, games in sorted(matchup_data, key=lambda x: (-x[2], x[1], -x[3], x[0])):
        print(f"{champion: >15}: {winrate: 5.2f}% | {games: >5} games | Priority: {priority}")
    
    if len(no_data) > 0:
        print("No data for the following champions:")
        for champion, priority in sorted(no_data, key=lambda x: (x[1], x[0])):
            print(f"{champion: >15} | Priority: {priority}")

def main():
    app()

if __name__ == "__main__":
    main()


import requests
import bs4
import re
from warnings import warn
from helpmepick.models.champion import Champion
from helpmepick.models.region import Region
from helpmepick.models.role import Role
from helpmepick.models.tier import Tier
from rapidfuzz.distance.DamerauLevenshtein import distance


def retrieve_opgg_data(
    champion: Champion, role: Role, region: Region = Region.g, tier: Tier = Tier.e_
) -> requests.models.Response:
    url = f"https://www.op.gg/champions/{champion.name}/counters/{role.name}"
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36"
    }
    params = {"region": region, "tier": tier}
    resp = requests.get(url, headers=headers, params=params)

    if resp.status_code != 200:
        raise ValueError(
            f"Request on url {resp.url} threw status code {resp.status_code}"
        )
    return resp


def parse_champion_name(name: str) -> Champion:
    plain = "".join([letter for letter in name.lower() if letter.isalpha()])
    # try to fit it plainly, else use a min-dist approach
    try:
        champion = Champion(plain)
    except ValueError:
        mindist = 100.0
        champion = Champion.annie
        for candidate in Champion:
            dist = distance(plain, candidate)
            if dist < mindist:
                mindist = dist
                champion = candidate
            # this shouldn't happen but oh well
            if dist == 0:
                break
    return champion


def parse_row(row: str) -> tuple[Champion, float, int]:
    pattern = r"(\D+)(\d+[.,]?\d*)%(\d+[.,]?\d+)"
    res = re.findall(pattern, row)
    if len(res) > 1:
        warn(f"Warning: parsing row {row} threw multiple result groups")
    name, winrate, matches = res[0]
    # matches number may contain a comma, e.g. '1,584'
    return parse_champion_name(name), float(winrate), int(matches.replace(",", ""))


def parse_data(resp: requests.models.Response) -> list[tuple[Champion, float, int]]:
    soup = bs4.BeautifulSoup(resp.content, "html.parser")
    table_list = soup.find_all("table")
    if len(table_list) != 1:
        warn(f"Warning: table list for url {resp.url} has multiple tables")
    # first row is omitted as it's the header
    rows = table_list[0].find_all("tr")[1:]
    return [parse_row(row.text) for row in rows]



from helpmepick.models.champion import Champion
from helpmepick.models.role import Role

RolePool = dict[Champion, int]


class Pool:
    def __init__(self, pool: dict[Role, RolePool]) -> None:
        if set(pool) != set(Role):
            raise ValueError(
                f"Passed pool is missing roles: {', '.join(set(Role) - set(pool))}"
            )
        self.pool = pool

    def _exists(self, champion: Champion, role: Role, isin: bool = True) -> None:
        """Raises an exception if champion does not exist for a role,
        or if it does if isin is False."""
        if (champion in self.pool[role]) ^ isin:
            raise ValueError(
                f"Champion {champion} {'not' if isin else 'already'} in pool for role {role}"
            )

    def add_champion(self, champion: Champion, role: Role, priority: int) -> None:
        self._exists(champion, role, isin=False)
        if champion in self.pool[role]:
            raise ValueError(f"Champion {champion} already in pool for role {role}")
        self.pool[role][champion] = priority

    def delete_champion(self, champion: Champion, role: Role) -> None:
        self._exists(champion, role)
        del self.pool[role][champion]

    def edit_champion(self, champion: Champion, role: Role, priority: int) -> None:
        self._exists(champion, role)
        self.pool[role][champion] = priority

    def get_champions_for(self, role: Role) -> RolePool:
        return self.pool[role]



from enum import Enum


class Region(str, Enum):
    g = "global"
    kr = "kr"
    na = "na"
    eu = "euw"
    las = "las"
    lan = "lan"
    br = "br"
    jp = "jp"
    oce = "oce"
    sg = "sg"
    vn = "vn"
    tw = "th"
    th = "th"
    eune = "eune"
    ru = "ru"



from pathlib import Path
import yaml
from helpmepick.models.champion import Champion
from helpmepick.models.pool import Pool
from helpmepick.models.role import Role
from helpmepick.pool.base import BasePoolIO


class YamlPoolIO(BasePoolIO):
    def __init__(self, path: Path) -> None:
        self.path = path

    def load_pool(self) -> Pool:
        assert self.path.exists(), f"Passed pool path {self.path} does not exist"
        with open(self.path) as f:
            pool = yaml.safe_load(f)
        return Pool(
            {
                Role(r): {Champion(c): prio for c, prio in self.orderedpool(rolepool)}
                for r, rolepool in pool.items()
            }
        )

    def save_pool(self, pool: Pool) -> None:
        serialized = {
            r.value: {c.name: prio for c, prio in self.orderedpool(rolepool)}
            for r, rolepool in pool.pool.items()
        }
        with open(self.path, "wt") as f:
            yaml.safe_dump(serialized, f, sort_keys=False)



from math import sqrt
import random as r
from time import sleep

AVAILABLE_WAIT_FUNCTIONS = {
    "normal": lambda m,s: r.gauss(m,s),
    "uniform": lambda m,s: r.uniform(m-sqrt(12)*s, m+sqrt(12)*s),
    "exp": lambda m, s: r.expovariate(1/m)
}


class Throttler:

    def __init__(
            self, 
            iterable, 
            mean: float = 3, 
            sd: float = 0.2, 
            distro: str="normal", 
            min_sleep: float = 1.0
    ) -> None:
        self.iterable = iterable
        self.wait_f = AVAILABLE_WAIT_FUNCTIONS[distro]
        self.mu = mean
        self.sigma = sd
        self.min_sleep = min_sleep

    def sleep(self):
        sleep(max(self.min_sleep, self.wait_f(self.mu, self.sigma)))

    def __iter__(self):
        for element in self.iterable:
            yield element
            self.sleep()


from abc import ABC, abstractmethod
from typing import Iterator
from helpmepick.models.pool import Pool, RolePool


class BasePoolIO(ABC):
    @abstractmethod
    def load_pool(self) -> Pool:
        pass

    @abstractmethod
    def save_pool(self, pool: Pool) -> None:
        pass

    def orderedpool(self, rolepool: RolePool) -> Iterator[tuple[str,int]]:
        """Iterate through a RolePool first sorted by prio, then champion name."""
        for k, v in sorted(rolepool.items(), key=lambda t: t[::-1]):
            yield k,v



from enum import Enum


class Champion(str, Enum):
    aatrox = "aatrox"
    ahri = "ahri"
    akali = "akali"
    akshan = "akshan"
    alistar = "alistar"
    amumu = "amumu"
    anivia = "anivia"
    annie = "annie"
    aphelios = "aphelios"
    ashe = "ashe"
    aurelionsol = "aurelionsol"
    azir = "azir"
    bard = "bard"
    belveth = "belveth"
    blitzcrank = "blitzcrank"
    brand = "brand"
    braum = "braum"
    briar = "briar"
    caitlyn = "caitlyn"
    camille = "camille"
    cassiopeia = "cassiopeia"
    chogath = "chogath"
    corki = "corki"
    darius = "darius"
    diana = "diana"
    drmundo = "drmundo"
    draven = "draven"
    ekko = "ekko"
    elise = "elise"
    evelynn = "evelynn"
    ezreal = "ezreal"
    fiddlesticks = "fiddlesticks"
    fiora = "fiora"
    fizz = "fizz"
    galio = "galio"
    gangplank = "gangplank"
    garen = "garen"
    gnar = "gnar"
    gragas = "gragas"
    graves = "graves"
    gwen = "gwen"
    hecarim = "hecarim"
    heimerdinger = "heimerdinger"
    illaoi = "illaoi"
    irelia = "irelia"
    ivern = "ivern"
    janna = "janna"
    jarvaniv = "jarvaniv"
    jax = "jax"
    jayce = "jayce"
    jhin = "jhin"
    jinx = "jinx"
    kaisa = "kaisa"
    kalista = "kalista"
    karma = "karma"
    karthus = "karthus"
    kassadin = "kassadin"
    katarina = "katarina"
    kayle = "kayle"
    kayn = "kayn"
    kennen = "kennen"
    khazix = "khazix"
    kindred = "kindred"
    kled = "kled"
    kogmaw = "kogmaw"
    ksante = "ksante"
    leblanc = "leblanc"
    leesin = "leesin"
    leona = "leona"
    lillia = "lillia"
    lissandra = "lissandra"
    lucian = "lucian"
    lulu = "lulu"
    lux = "lux"
    malphite = "malphite"
    malzahar = "malzahar"
    maokai = "maokai"
    masteryi = "masteryi"
    millio = "millio"
    missfortune = "missfortune"
    mordekaiser = "mordekaiser"
    morgana = "morgana"
    naafiri = "naafiri"
    nami = "nami"
    nasus = "nasus"
    nautilus = "nautilus"
    neeko = "neeko"
    nidalee = "nidalee"
    nilah = "nilah"
    nocturne = "nocturne"
    nunu = "nunu"
    olaf = "olaf"
    orianna = "orianna"
    ornn = "ornn"
    pantheon = "pantheon"
    poppy = "poppy"
    pyke = "pyke"
    qiyana = "qiyana"
    quinn = "quinn"
    rakan = "rakan"
    rammus = "rammus"
    reksai = "reksai"
    rell = "rell"
    renata = "renata"
    renekton = "renekton"
    rengar = "rengar"
    riven = "riven"
    rumble = "rumble"
    ryze = "ryze"
    samira = "samira"
    sejuani = "sejuani"
    senna = "senna"
    seraphine = "seraphine"
    sett = "sett"
    shaco = "shaco"
    shen = "shen"
    shyvana = "shyvana"
    singed = "singed"
    sion = "sion"
    sivir = "sivir"
    skarner = "skarner"
    sona = "sona"
    soraka = "soraka"
    swain = "swain"
    sylas = "sylas"
    syndra = "syndra"
    tahmkench = "tahmkench"
    taliyah = "taliyah"
    talon = "talon"
    taric = "taric"
    teemo = "teemo"
    thresh = "thresh"
    tristana = "tristana"
    trundle = "trundle"
    tryndamere = "tryndamere"
    twistedfate = "twistedfate"
    twitch = "twitch"
    udyr = "udyr"
    urgot = "urgot"
    varus = "varus"
    vayne = "vayne"
    veigar = "veigar"
    velkoz = "velkoz"
    vex = "vex"
    vi = "vi"
    viego = "viego"
    viktor = "viktor"
    vladimir = "vladimir"
    volibear = "volibear"
    warwick = "warwick"
    wukong = "wukong"
    xayah = "xayah"
    xerath = "xerath"
    xinzhao = "xinzhao"
    yasuo = "yasuo"
    yone = "yone"
    yorick = "yorick"
    yuumi = "yuumi"
    zac = "zac"
    zed = "zed"
    zeri = "zeri"
    ziggs = "ziggs"
    zilean = "zilean"
    zoe = "zoe"
    zyra = "zyra"



import csv
from pathlib import Path
from helpmepick.models.champion import Champion
from helpmepick.persistence.base import BaseChampIO


class CsvChampIO(BaseChampIO):
    def __init__(self, data_folder: Path) -> None:
        super().__init__(data_folder)

    def get_champ_path(self, champion: Champion) -> Path:
        return self.root / f"{champion}.csv"

    def load_champ(self, champion: Champion) -> list[tuple[Champion, float, int]]:
        with open(self.get_champ_path(champion)) as f:
            r = csv.reader(f)
            res = [
                (Champion(champion), float(wr), int(matches))
                for champion, wr, matches in r
            ]
        return res

    def save_champ(
        self, champion: Champion, data: list[tuple[Champion, float, int]]
    ) -> None:
        with open(self.get_champ_path(champion), "wt", newline='') as f:
            w = csv.writer(f)
            w.writerows(data)
        return
